@@ -75,7 +75,7 @@ from types import GeneratorType, ModuleType
 from collections import deque
 from binascii import a2b_hex, b2a_hex
 
-VERSION = '0.02'
+VERSION = '0.03'
 
 PYTHON3 = sys.version_info >= (3, 0)
 if PYTHON3:
@@ -114,9 +114,9 @@ __all__ = ['AutoThreadingClient', 'Client', 'ClientGroup', 'Connection',
            'RemoteProxy', 'SelfIterProxy', 'Server', 'ServerObject',
            'add_key_options', 'call_method_multi', 'call_multi', 'decode_key',
            'define_common', 'definitions_module', 'eval_multi', 'exec_return',
-           'execblock_multi', 'key_from_options', 'make_server_class',
-           'make_singleton', 'read_key_file', 'run_server',
-           'server_check_runner', 'spawn_local']
+           'execblock_multi', 'generate_key', 'key_from_options',
+           'make_server_class', 'make_singleton', 'read_key_file', 'run_server',
+           'server_check_runner', 'spawn_local', 'write_key_file',]
 
 ## Special value for TID which causes thread to exit after one request
 SINGLE = -1
@@ -277,6 +277,14 @@ def make_singleton(obj, export_methods=(), export_attrs=(), proxy_class=None):
         
     _singletons[addr] = export_methods, export_attrs, proxy_class, wr, obj
 
+class DummyConnection(object):
+    '''Replacement object for _connection in invalidated proxies.'''
+    peername = '<closed connection>'
+    def send_request(self, *a, **kw):
+        raise IOError("Connection has been closed")
+
+_dummy_connection = DummyConnection()
+
 class RemoteProxyRef(weakref.ref):
     __slots__ = ('oid',)
     def __new__(cls, prox, cb, oid):
@@ -332,6 +340,20 @@ class RemoteProxy(RemoteProxyBase):
         self._connection = client
         self._oid = oid
 
+    @property
+    def proxy_connection(self):
+        '''Returns the connection that this object is associated with, or None
+        if the connection was closed.'''
+        ret = self._connection
+        if ret is _dummy_connection:
+            ret = None
+        return ret
+
+    @property
+    def proxy_object_id(self):
+        '''The object ID assigned by the remote connection for this object'''
+        return self._oid
+
     def raw_call_method(self, method, args, kwargs={}, nonblocking=False,
                         oncomplete=None, onerror=None, threadid=None):
         '''Call a method on the remote object. 
@@ -440,30 +462,58 @@ class Connection:
 
         self._sock = None
         self._peername = ''
+        self._remote_host = None
+        self._remote_port = None
         
         self._lock = threading.Lock()
         self._sendlock = threading.Lock()
 
-        self._recv_hmac = hmac.new(key, self._recviv, HMAC_MOD)
-        self._send_hmac = hmac.new(key, self._sendiv, HMAC_MOD)
-
     @property
     def peername(self):
         '''The name of the host we are connected to'''
         return self._peername
 
+    @property
+    def connected(self):
+        '''True if we are connected to a remote host.'''
+        return self._sock is not None
+
+    @property
+    def remote_host(self):
+        '''The hostname or IP address passed to connect(), or None for a local
+        connection'''
+        return self._remote_host
+
+    @property
+    def remote_port(self):
+        '''The TCP port passed to connect(), or None for a local connection'''
+        return self._remote_port
+
     def connect(self, host, port=DEFAULT_PORT):
         '''Connect to a remote host.'''
+        self._remote_host = host
+        self._remote_port = port
         self._peername = '%s:%d' % (host, port)
         self.set_socket(socket.create_connection((host, port)))
         
+    def reconnect(self):
+        '''Reconnect to the remote host. This will reset the remote environment,
+        and any currently held proxied objects will become invalid.'''
+        if self._remote_host is None:
+            raise ValueError('Cannot reconnect before first connection.')
+        self.close()
+        self.connect(self._remote_host, self._remote_port)
+
     def set_socket(self, sock, peername=None):
         '''Use an already-open socket for this connection.'''
         if peername is None:
             peername = _get_peer_name(sock.getpeername())
+
         _mark_non_inheritable(sock.fileno())
         self._sock = sock
         self._peername = peername
+        self._recv_hmac = hmac.new(self._key, self._recviv, HMAC_MOD)
+        self._send_hmac = hmac.new(self._key, self._sendiv, HMAC_MOD)
         self._after_connect()
         
     def _after_connect(self):
@@ -581,8 +631,9 @@ class Connection:
 
     def close(self):
         '''Close the connection.'''
-        self._sock.shutdown(socket.SHUT_RDWR)
-        self._sock = None
+        if self._sock is not None:
+            self._sock.shutdown(socket.SHUT_RDWR)
+            self._sock = None
 
 class _GlobalProxy:
     '''A proxy for a global object on a remote server'''
@@ -631,7 +682,7 @@ class Client(Connection):
         Connection.__init__(self, key)
 
         self._next_request_id = 0
-        self._pending_requests = []
+        self._pending_requests = None
         self._pending_responses = {}
         self._completion = {}
 
@@ -640,7 +691,7 @@ class Client(Connection):
         def proxy_deleted(refdat, wrself=weakref.ref(self)):
             self = wrself()
             if self is not None:
-                del self._proxy_by_id[refdat.oid]
+                self._proxy_by_id.pop(refdat.oid, None)
                 self._garbage_proxies.append(refdat.oid)
 
         self._proxy_deleted = proxy_deleted
@@ -665,6 +716,7 @@ class Client(Connection):
 
     def _after_connect(self):
         self._is_ready = False
+        self._pending_requests = []
         self._sock.setblocking(0)
         self._init_thread()
         self.send_challenge()
@@ -763,6 +815,8 @@ class Client(Connection):
         See NonblockingResponse.wait for details on exception handling.
         '''
         with self._lock:
+            if self._sock is None:
+                raise IOError("Not connected")
             if threadid is None:
                 threadid = self._thread_id()
 
@@ -780,8 +834,7 @@ class Client(Connection):
             except Exception:
                 gobj.extend(garbage)
                 raise
-            v = self._is_ready, self._pending_requests
-            assert (v[0] is False) == (v[1] is not None)
+
             if not self._is_ready:
                 self._pending_requests.append(rq)
                 rq = None
@@ -982,6 +1035,16 @@ class Client(Connection):
         self._proxy_by_id[oid] = ref
         return prox
             
+    def close(self):
+        '''Close the connection and invalidate all remote proxies.'''
+        Connection.close(self)
+        proxies = self._proxy_by_id
+        while proxies:
+            oid, wrprox = proxies.popitem()
+            prox = wrprox()
+            if prox is not None:
+                prox._connection = _dummy_connection
+
 class AutoThreadingClient(Client, threading.Thread):
     '''A client that automatically creates unique threads on the server for each
     thread on the client that makes remote calls.
@@ -1417,6 +1480,21 @@ def read_key_file(path):
     with open(path, 'r') as fp:
         return decode_key(fp.readline().rstrip('\r\n'))
 
+def write_key_file(path, key, overwrite=False):
+    '''Write an HMAC key to a file. If overwrite is True, it will try to
+    overwrite the key file if it already exists.'''
+    if not _write_key_file(path, key):
+        if not overwrite:
+            return False
+        os.unlink(path)
+        return _write_key_file(path, key)
+    return True
+
+def generate_key():
+    '''Generates a random HMAC key of the proper size.'''
+    return os.urandom(HMAC_SIZE)
+
+
 def add_key_options(options, key_option='-K', keyfile_option='-k'):
     '''Add options for specifying a HMAC key or keyfile to an optparse.OptionParser.'''
     options.add_option(key_option, '--key', help='HMAC key to use; either a %d-character hex string '
@@ -1465,21 +1543,10 @@ def server_main(args=None):
     opts, args = options.parse_args(args)
 
     if opts.generate:
-        try:
-            os.umask(0o177)
-        except Exception:
-            pass
-        key = os.urandom(HMAC_SIZE)
-        if not opts.overwrite and os.path.exists(opts.generate):
+        if not write_key_file(opts.generate, generate_key(), opts.overwrite):
             print('Key file already exists (use --overwrite to overwrite it)', file=sys.stderr)
-            return
-        with open(opts.generate, 'w') as fp:
-            try:
-                os.fchmod(fp.fileno(), 0o600)
-            except Exception:
-                pass
-            fp.write(_hexenc(key) + '\n')
         return
+
     elif opts.serve:
         key = key_from_options(opts)
         if '.' in opts.svrclass:
@@ -1832,6 +1899,37 @@ if os.name == 'nt':
     def server_init():
         pass
 
+    def _write_key_file(path, key):
+        '''Write an HMAC key to a file. Returns True if the file was created, or
+        False if it already exists.'''
+        import win32api, win32file, win32security
+        import ntsecuritycon, win32con, pywintypes
+        user, domain, type = win32security.LookupAccountName ("", win32api.GetUserName())
+
+        dacl = win32security.ACL()
+        dacl.AddAccessAllowedAce(win32security.ACL_REVISION, 
+                                 ntsecuritycon.FILE_ALL_ACCESS, user)
+
+        sd = win32security.SECURITY_DESCRIPTOR()
+        sd.SetSecurityDescriptorDacl(1, dacl, 0)
+
+        sa = win32security.SECURITY_ATTRIBUTES()
+        sa.SECURITY_DESCRIPTOR = sd
+        sa.bInheritHandle = 0
+
+        try:
+            fil = win32file.CreateFile(path, win32con.GENERIC_WRITE, 0, sa, 
+                                       win32con.CREATE_NEW, 0, 0)
+        except pywintypes.error as e:
+            ## File exists.
+            if e[0] == 80:
+                return False
+        try:
+            win32file.WriteFile(fil, _hexenc(key) + '\r\n', None)
+        finally:
+            win32file.CloseHandle(fil)
+        return True
+
 else:
     import fcntl
     
@@ -1870,5 +1968,19 @@ else:
     def server_init():
         signal.signal(signal.SIGCLD, sigcld)
 
+    def _write_key_file(path, key):
+        '''Write an HMAC key to a file. Returns True if the file was created, or
+        False if it already exists.'''
+        try:
+            fd = os.open(path, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0o600)
+        except OSError as e:
+            if e[0] == errno.EEXIST:
+                return False
+        try:
+            os.write(fd, _hexenc(key) + '\n')
+        finally:
+            os.close(fd)
+        return True
+
 if _is_main:
     server_main()
